/*
 * Copyright (c) 1999 G. Adam Stanislav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	ptrans.c v.1, Unix version
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <utf-8.h>

/*
 * Note: If you do not have <utf-8.h>, download libutf-8-1.0.tar.gz from
 * http://www.whizkidtech.net/i18n/
 *
 * The library enclosed in that package (libutf-8.so) is necessary for ptrans
 * to work.
 */

static int verbose  = 0;
static int binary   = 1;
static int eightbit = 1;
static unsigned int line      = 1;
static unsigned int character = 0;
static FILE *input   = stdin;
static FILE *output  = stdout;
static FILE *page    = NULL;
static char *charmap = NULL;
static char buffer[9];
static int uni[256] = {
	0,
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	9,
	10,
	11,
	12,
	13,
	14,
	15,
	16,
	17,
	18,
	19,
	20,
	21,
	22,
	23,
	24,
	25,
	26,
	27,
	28,
	29,
	30,
	31,
	32,
	33,
	34,
	35,
	36,
	37,
	38,
	39,
	40,
	41,
	42,
	43,
	44,
	45,
	46,
	47,
	48,
	49,
	50,
	51,
	52,
	53,
	54,
	55,
	56,
	57,
	58,
	59,
	60,
	61,
	62,
	63,
	64,
	65,
	66,
	67,
	68,
	69,
	70,
	71,
	72,
	73,
	74,
	75,
	76,
	77,
	78,
	79,
	80,
	81,
	82,
	83,
	84,
	85,
	86,
	87,
	88,
	89,
	90,
	91,
	92,
	93,
	94,
	95,
	96,
	97,
	98,
	99,
	100,
	101,
	102,
	103,
	104,
	105,
	106,
	107,
	108,
	109,
	110,
	111,
	112,
	113,
	114,
	115,
	116,
	117,
	118,
	119,
	120,
	121,
	122,
	123,
	124,
	125,
	126,
	127,
	128,
	129,
	130,
	131,
	132,
	133,
	134,
	135,
	136,
	137,
	138,
	139,
	140,
	141,
	142,
	143,
	144,
	145,
	146,
	147,
	148,
	149,
	150,
	151,
	152,
	153,
	154,
	155,
	156,
	157,
	158,
	159,
	160,
	161,
	162,
	163,
	164,
	165,
	166,
	167,
	168,
	169,
	170,
	171,
	172,
	173,
	174,
	175,
	176,
	177,
	178,
	179,
	180,
	181,
	182,
	183,
	184,
	185,
	186,
	187,
	188,
	189,
	190,
	191,
	192,
	193,
	194,
	195,
	196,
	197,
	198,
	199,
	200,
	201,
	202,
	203,
	204,
	205,
	206,
	207,
	208,
	209,
	210,
	211,
	212,
	213,
	214,
	215,
	216,
	217,
	218,
	219,
	220,
	221,
	222,
	223,
	224,
	225,
	226,
	227,
	228,
	229,
	230,
	231,
	232,
	233,
	234,
	235,
	236,
	237,
	238,
	239,
	240,
	241,
	242,
	243,
	244,
	245,
	246,
	247,
	248,
	249,
	250,
	251,
	252,
	253,
	254,
	255
};

static int uniind[256] = {
	0,
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	9,
	10,
	11,
	12,
	13,
	14,
	15,
	16,
	17,
	18,
	19,
	20,
	21,
	22,
	23,
	24,
	25,
	26,
	27,
	28,
	29,
	30,
	31,
	32,
	33,
	34,
	35,
	36,
	37,
	38,
	39,
	40,
	41,
	42,
	43,
	44,
	45,
	46,
	47,
	48,
	49,
	50,
	51,
	52,
	53,
	54,
	55,
	56,
	57,
	58,
	59,
	60,
	61,
	62,
	63,
	64,
	65,
	66,
	67,
	68,
	69,
	70,
	71,
	72,
	73,
	74,
	75,
	76,
	77,
	78,
	79,
	80,
	81,
	82,
	83,
	84,
	85,
	86,
	87,
	88,
	89,
	90,
	91,
	92,
	93,
	94,
	95,
	96,
	97,
	98,
	99,
	100,
	101,
	102,
	103,
	104,
	105,
	106,
	107,
	108,
	109,
	110,
	111,
	112,
	113,
	114,
	115,
	116,
	117,
	118,
	119,
	120,
	121,
	122,
	123,
	124,
	125,
	126,
	127,
	128,
	129,
	130,
	131,
	132,
	133,
	134,
	135,
	136,
	137,
	138,
	139,
	140,
	141,
	142,
	143,
	144,
	145,
	146,
	147,
	148,
	149,
	150,
	151,
	152,
	153,
	154,
	155,
	156,
	157,
	158,
	159,
	160,
	161,
	162,
	163,
	164,
	165,
	166,
	167,
	168,
	169,
	170,
	171,
	172,
	173,
	174,
	175,
	176,
	177,
	178,
	179,
	180,
	181,
	182,
	183,
	184,
	185,
	186,
	187,
	188,
	189,
	190,
	191,
	192,
	193,
	194,
	195,
	196,
	197,
	198,
	199,
	200,
	201,
	202,
	203,
	204,
	205,
	206,
	207,
	208,
	209,
	210,
	211,
	212,
	213,
	214,
	215,
	216,
	217,
	218,
	219,
	220,
	221,
	222,
	223,
	224,
	225,
	226,
	227,
	228,
	229,
	230,
	231,
	232,
	233,
	234,
	235,
	236,
	237,
	238,
	239,
	240,
	241,
	242,
	243,
	244,
	245,
	246,
	247,
	248,
	249,
	250,
	251,
	252,
	253,
	254,
	255
};

static int usage() {
	fprintf(stderr, "Usage: ptrans [-p pagefile][-i inputfile][-o outputfile][-v][-e][-b|t][-7|8]\n");
	return 1;
}

static int cantopen(char *file) {
	fprintf(stderr, "ptrans: Can't open %s.\n", file);
	if (charmap != NULL)
		free(charmap);
	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);
	/* No need to close page file, it is always NULL at this point */
	return 2;
}

static int nem(void) {
	fprintf(stderr, "ptrans: Not enough memory.\n");
	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);
	return 3;
}

static void unisort(void) {
	int i, j;
	int sorted;
	int temp;

	/*
	 * This sort is optimized for two assumptions:
	 *
	 * (1) The bottom half of the table is already sorted;
	 * (2) The top half is near sorted.
	 *
	 * That leads to the use of two bubble sorts (which
	 * perform well on near sorted tables).
	 *
	 * The first one sorts the top half of the table,
	 * presumbly in fewer steps than quick sort would.
	 * (Quick sort performs poorly on near sorted tables.)
	 *
	 * The second one will most likely just go through one
	 * pass, not having to do anything.
	 *
	 * Our two assumptions perform well with various ISO-8859
	 * encodings, and poorly with something like EBCDIC.
	 */

	/* First, sort the upper half only */
	for (j = 255; j > 127; j--) {
		sorted = 1;

		for (i = 127; i < j; i++) {
			if (uni[uniind[i]] > uni[uniind[i+1]]) {
				sorted      = 0;
				temp        = uniind[i];
				uniind[i]   = uniind[i+1];
				uniind[i+1] = temp;
			}
		}

		if (sorted) break;
	}

	/* Second, sort the whole table if necessary. */
	for (j = 255; j; j--) {
		sorted = 1;

		for(i = 0; i < j; i++) {
			if (uni[uniind[i]] > uni[uniind[i+1]]) {
				sorted      = 0;
				temp        = uniind[i];
				uniind[i]   = uniind[i+1];
				uniind[i+1] = temp;
			}
		}
		if (sorted) break;
	}
}

static int unisearch(int c) {
	int min = 0;
	int max = 255;
	int middle;

	for (;;) {
		middle = (min + max) / 2;

		if (uni[uniind[middle]] == c)
			return uniind[middle];

		else if (min == max)
			return -1;

		else if (uni[uniind[middle]] < c)
			min = middle + 1;

		else max = middle - 1;
	}
}

static void parse(void) {
	int c;
	int chars;
	unsigned int pagechar;
	unsigned int unichar;

	if (binary != 0) {
		if (verbose)
			fprintf(stderr, "ptrans: Loading binary charmap.\n");
		fread(uni, sizeof(unsigned int), 256, page);
		unisort();
		return;
	}

	c = 0;

	while (c != EOF) {
		c = fgetc(page);
		if ((c == '=') || ((c == '/') && (fgetc(page) == 'x'))) {
			for (chars = 0; chars < 2; chars++) {
				c = fgetc(page);
				if (isxdigit(c))
					buffer[chars] = c;
				else {
					ungetc(c, page);
					break;
				}
			}
			buffer[chars] = '\0';
			pagechar = strtoul(buffer, NULL, 16);
			for (;;) {
				c = fgetc(page);
				if (c == EOF) {
					unisort();
					return;
				}
				else if (c == 'U') {
					c = fgetc(page);
					if (c != '+') {
						ungetc(c, page);
						if (!isxdigit(c)) continue;
					}

					for (chars = 0; chars < 8; chars++) {
						c = fgetc(page);
						if (isxdigit(c))
							buffer[chars] = c;
						else {
							ungetc(c, page);
							break;
						}
					}
					buffer[chars] = '\0';
					unichar = strtoul(buffer, NULL, 16);
					if (pagechar < 256) {
						uni[pagechar] = unichar;
						if (verbose)
							fprintf(stderr, "ptrans: Converting %02x to %04x.\n", pagechar, unichar);
					}
					else if (verbose)
						fprintf(stderr, "ptrans: Ignoring invalid character %02x.\n", pagechar);
					break;
				}
			}
		}
	}
	unisort();
}

int main(int argc, char *argv[]) {
	int i, j, k;
	int ignoreenvironment = 0;
	char *inp  = NULL;
	char *outp = NULL;
	char *pg   = NULL;
	char *cp;

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			return usage();
		switch (argv[i][1]) {
			case 'b':
				binary = 1;
				break;
			case 't':
				binary = 0;
				break;
			case 'p':
				if (++i < argc)
					pg = argv[i];
				else return usage();
				break;
			case 'i':
				if (++i < argc)
					inp = argv[i];
				else return usage();
				break;
			case 'o':
				if (++i < argc)
					outp = argv[i];
				else return usage();
				break;
			case 'e':
				ignoreenvironment = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case '7':
				eightbit = 0;
				break;
			case '8':
				eightbit = 1;
				break;
			default:
				return usage();
		}
	}

	cp = ignoreenvironment == 0 ? getenv("CHARMAPS") : NULL;

	if ((pg == NULL) && (ignoreenvironment == 0) && ((pg = getenv("PTRANS")) == NULL))
		pg = getenv("UTRANS");

	if (inp != NULL) {
		input = fopen(inp, "r");
		if (input == NULL)
			return cantopen(inp);
	}

	if (outp != NULL) {
		output = fopen(outp, "w");
		if (output == NULL)
			return cantopen(outp);
	}

	if (pg != NULL) {
		if (cp == NULL) {
			charmap = strdup(pg);
			if (charmap == NULL)
				return nem();
		}
		else {
			charmap = (char *)malloc(sizeof(pg) + sizeof(cp) + 2);
			if (charmap == NULL)
				return nem();
			sprintf(charmap, "%s/%s", cp, pg);
		}
		page = fopen(charmap, "r");
		if (page == NULL)
			return cantopen(charmap);
		else {
			free(charmap);
			charmap = NULL;
		}
	}

	if (page != NULL) {
		parse();
		fclose(page);
	}

	while ((i = utf8fgetc(input)) != EOF) {
		if (i == '\n') {
			line++;
			character = 0;
		}
		character++;
		k = unisearch(i);
		if ((k >= 0) && (eightbit || (k < 0x80))) {
			j = fputc(k, output);
			if (j == EOF) {
				fprintf(stderr, "ptrans (line %u, character %u): Error writing output (%02x, U+%04x, %c).\n", line, character, i, uni[k], i);
				break;
			}
		}
		else
			fprintf(output, "&#%i;", i);
	}

	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);

	return 0;
}

