/*
 * Copyright (c) 1999 G. Adam Stanislav
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	uhtrans.c v.1, Unix version
 */

#include <stdio.h>
#include <utf-8.h>

/*
 * Note: If you do not have <utf-8.h>, download libutf-8-1.0.tar.gz from
 * http://www.whizkidtech.net/i18n/
 *
 * The library enclosed in that package (libutf-8.so) is necessary for uhtrans
 * to work.
 */

#define MAX_ERRORS	8	/* Arbitrary value, change if you want */

static unsigned int line = 1;
static unsigned int numerrors = 0;
static FILE *input   = stdin;
static FILE *output  = stdout;

int usage() {
	fprintf(stderr, "Usage: uhtrans [ -i inputfile][ -o outputfile]\n");
	return 1;
}

int cantopen(char *file) {
	fprintf(stderr, "uhtrans: Can't open %s.\n", file);
	if ((input != NULL) && (input != stdin))
		fclose(input);
	if ((output != NULL) && (output != stdout))
		fclose(output);
	return 2;
}

int main(int argc, char *argv[]) {
	int i;
	unsigned int c;
	char *inp  = NULL;
	char *outp = NULL;

	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			return usage();
		switch (argv[i][1]) {
			case 'i':
				if (++i < argc)
					inp = argv[i];
				else return usage();
				break;
			case 'o':
				if (++i < argc)
					outp = argv[i];
				else return usage();
				break;
			default:
				return usage();
		}
	}

	if (inp != NULL) {
		input = fopen(inp, "r");
		if (input == NULL)
			return cantopen(inp);
	}

	if (outp != NULL) {
		output = fopen(outp, "w");
		if (output == NULL)
			return cantopen(outp);
	}

	for (;;) {
		c = utf8fgetc(input);

		if (c == EOF)
			break;
		else if (c >= 0x80000000) {
			fprintf(stderr, "uhtrans: Invalid UTC-8 sequence in line %u.\n", line);
			if (++numerrors >= MAX_ERRORS) {
				/* Assume file not UTC-8 encoded */
				fprintf(stderr, "uhtrans: Too many invalid UTC-8 sequences. Aborting.\n");
				break;
			}
			else
				continue;
		}

		if (c == '\n') line++;

		if (c < 0x80)
			fputc(c, output);
		else
			fprintf(output, "&#%u;", c);
	}

	if (input != stdin)
		fclose(input);

	if (output != stdout)
		fclose(output);

	return numerrors;
}
